/**

seedrandom.js
=============

Seeded random number generator for Javascript.

version 2.3.10
Author: David Bau
Date: 2014 Sep 20

Can be used as a plain script, a node.js module or an AMD module.

Script tag usage
----------------

<script src=//cdnjs.cloudflare.com/ajax/libs/seedrandom/2.3.10/seedrandom.min.js>
</script>

// Sets Math.random to a PRNG initialized using the given explicit seed.
Math.seedrandom('hello.');
console.log(Math.random());          // Always 0.9282578795792454
console.log(Math.random());          // Always 0.3752569768646784

// Sets Math.random to an ARC4-based PRNG that is autoseeded using the
// current time, dom state, and other accumulated local entropy.
// The generated seed string is returned.
Math.seedrandom();
console.log(Math.random());          // Reasonably unpredictable.

// Seeds using the given explicit seed mixed with accumulated entropy.
Math.seedrandom('added entropy.', { entropy: true });
console.log(Math.random());          // As unpredictable as added entropy.

// Use "new" to create a local prng without altering Math.random.
var myrng = new Math.seedrandom('hello.');
console.log(myrng());                // Always 0.9282578795792454


Node.js usage
-------------

npm install seedrandom

// Local PRNG: does not affect Math.random.
var seedrandom = require('seedrandom');
var rng = seedrandom('hello.');
console.log(rng());                  // Always 0.9282578795792454

// Autoseeded ARC4-based PRNG.
rng = seedrandom();
console.log(rng());                  // Reasonably unpredictable.

// Global PRNG: set Math.random.
seedrandom('hello.', { global: true });
console.log(Math.random());          // Always 0.9282578795792454

// Mixing accumulated entropy.
rng = seedrandom('added entropy.', { entropy: true });
console.log(rng());                  // As unpredictable as added entropy.


Require.js usage
----------------

Similar to node.js usage:

bower install seedrandom

require(['seedrandom'], function(seedrandom) {
  var rng = seedrandom('hello.');
  console.log(rng());                  // Always 0.9282578795792454
});


Network seeding
---------------

<script src=//cdnjs.cloudflare.com/ajax/libs/seedrandom/2.3.10/seedrandom.min.js>
</script>

<!-- Seeds using urandom bits from a server. -->
<script src=//jsonlib.appspot.com/urandom?callback=Math.seedrandom">
</script>

<!-- Seeds mixing in random.org bits -->
<script>
(function(x, u, s){
  try {
    // Make a synchronous request to random.org.
    x.open('GET', u, false);
    x.send();
    s = unescape(x.response.trim().replace(/^|\s/g, '%'));
  } finally {
    // Seed with the response, or autoseed on failure.
    Math.seedrandom(s, !!s);
  }
})(new XMLHttpRequest, 'https://www.random.org/integers/' +
  '?num=256&min=0&max=255&col=1&base=16&format=plain&rnd=new');
</script>

Reseeding using user input
--------------------------

var seed = Math.seedrandom();        // Use prng with an automatic seed.
document.write(Math.random());       // Pretty much unpredictable x.

var rng = new Math.seedrandom(seed); // A new prng with the same seed.
document.write(rng());               // Repeat the 'unpredictable' x.

function reseed(event, count) {      // Define a custom entropy collector.
  var t = [];
  function w(e) {
    t.push([e.pageX, e.pageY, +new Date]);
    if (t.length &lt; count) { return; }
    document.removeEventListener(event, w);
    Math.seedrandom(t, { entropy: true });
  }
  document.addEventListener(event, w);
}
reseed('mousemove', 100);            // Reseed after 100 mouse moves.

The "pass" option can be used to get both the prng and the seed.
The following returns both an autoseeded prng and the seed as an object,
without mutating Math.random:

var obj = Math.seedrandom(null, { pass: function(prng, seed) {
  return { random: prng, seed: seed };
}});


Version notes
-------------

The random number sequence is the same as version 1.0 for string seeds.
* Version 2.0 changed the sequence for non-string seeds.
* Version 2.1 speeds seeding and uses window.crypto to autoseed if present.
* Version 2.2 alters non-crypto autoseeding to sweep up entropy from plugins.
* Version 2.3 adds support for "new", module loading, and a null seed arg.
* Version 2.3.1 adds a build environment, module packaging, and tests.
* Version 2.3.4 fixes bugs on IE8, and switches to MIT license.
* Version 2.3.6 adds a readable options object argument.
* Version 2.3.10 adds support for node.js crypto (contributed by ctd1500).

The standard ARC4 key scheduler cycles short keys, which means that
seedrandom('ab') is equivalent to seedrandom('abab') and 'ababab'.
Therefore it is a good idea to add a terminator to avoid trivial
equivalences on short string seeds, e.g., Math.seedrandom(str + '\0').
Starting with version 2.0, a terminator is added automatically for
non-string seeds, so seeding with the number 111 is the same as seeding
with '111\0'.

When seedrandom() is called with zero args or a null seed, it uses a
seed drawn from the browser crypto object if present.  If there is no
crypto support, seedrandom() uses the current time, the native rng,
and a walk of several DOM objects to collect a few bits of entropy.

Each time the one- or two-argument forms of seedrandom are called,
entropy from the passed seed is accumulated in a pool to help generate
future seeds for the zero- and two-argument forms of seedrandom.

On speed - This javascript implementation of Math.random() is several
times slower than the built-in Math.random() because it is not native
code, but that is typically fast enough.  Some details (timings on
Chrome 25 on a 2010 vintage macbook):

* seeded Math.random()          - avg less than 0.0002 milliseconds per call
* seedrandom('explicit.')       - avg less than 0.2 milliseconds per call
* seedrandom('explicit.', true) - avg less than 0.2 milliseconds per call
* seedrandom() with crypto      - avg less than 0.2 milliseconds per call

Autoseeding without crypto is somewhat slower, about 20-30 milliseconds on
a 2012 windows 7 1.5ghz i5 laptop, as seen on Firefox 19, IE 10, and Opera.
Seeded rng calls themselves are fast across these browsers, with slowest
numbers on Opera at about 0.0005 ms per seeded Math.random().


LICENSE (MIT)
-------------

Copyright 2014 David Bau.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

/**
 * All code is in an anonymous closure to keep the global namespace clean.
 */
(function (
    global, pool, math, width, chunks, digits, module, define, rngname) {

//
// The following constants are related to IEEE 754 limits.
//
var startdenom = math.pow(width, chunks),
    significance = math.pow(2, digits),
    overflow = significance * 2,
    mask = width - 1,
    nodecrypto;

//
// seedrandom()
// This is the seedrandom function described above.
//
var impl = math['seed' + rngname] = function(seed, options, callback) {
  var key = [];
  options = (options == true) ? { entropy: true } : (options || {});

  // Flatten the seed string or build one from local entropy if needed.
  var shortseed = mixkey(flatten(
    options.entropy ? [seed, tostring(pool)] :
    (seed == null) ? autoseed() : seed, 3), key);

  // Use the seed to initialize an ARC4 generator.
  var arc4 = new ARC4(key);

  // Mix the randomness into accumulated entropy.
  mixkey(tostring(arc4.S), pool);

  // Calling convention: what to return as a function of prng, seed, is_math.
  return (options.pass || callback ||
      // If called as a method of Math (Math.seedrandom()), mutate Math.random
      // because that is how seedrandom.js has worked since v1.0.  Otherwise,
      // it is a newer calling convention, so return the prng directly.
      function(prng, seed, is_math_call) {
        if (is_math_call) { math[rngname] = prng; return seed; }
        else return prng;
      })(

  // This function returns a random double in [0, 1) that contains
  // randomness in every bit of the mantissa of the IEEE 754 value.
  function() {
    var n = arc4.g(chunks),             // Start with a numerator n < 2 ^ 48
        d = startdenom,                 //   and denominator d = 2 ^ 48.
        x = 0;                          //   and no 'extra last byte'.
    while (n < significance) {          // Fill up all significant digits by
      n = (n + x) * width;              //   shifting numerator and
      d *= width;                       //   denominator and generating a
      x = arc4.g(1);                    //   new least-significant-byte.
    }
    while (n >= overflow) {             // To avoid rounding up, before adding
      n /= 2;                           //   last byte, shift everything
      d /= 2;                           //   right using integer math until
      x >>>= 1;                         //   we have exactly the desired bits.
    }
    return (n + x) / d;                 // Form the number within [0, 1).
  }, shortseed, 'global' in options ? options.global : (this == math));
};

//
// ARC4
//
// An ARC4 implementation.  The constructor takes a key in the form of
// an array of at most (width) integers that should be 0 <= x < (width).
//
// The g(count) method returns a pseudorandom integer that concatenates
// the next (count) outputs from ARC4.  Its return value is a number x
// that is in the range 0 <= x < (width ^ count).
//
/** @constructor */
function ARC4(key) {
  var t, keylen = key.length,
      me = this, i = 0, j = me.i = me.j = 0, s = me.S = [];

  // The empty key [] is treated as [0].
  if (!keylen) { key = [keylen++]; }

  // Set up S using the standard key scheduling algorithm.
  while (i < width) {
    s[i] = i++;
  }
  for (i = 0; i < width; i++) {
    s[i] = s[j = mask & (j + key[i % keylen] + (t = s[i]))];
    s[j] = t;
  }

  // The "g" method returns the next (count) outputs as one number.
  (me.g = function(count) {
    // Using instance members instead of closure state nearly doubles speed.
    var t, r = 0,
        i = me.i, j = me.j, s = me.S;
    while (count--) {
      t = s[i = mask & (i + 1)];
      r = r * width + s[mask & ((s[i] = s[j = mask & (j + t)]) + (s[j] = t))];
    }
    me.i = i; me.j = j;
    return r;
    // For robust unpredictability, the function call below automatically
    // discards an initial batch of values.  This is called RC4-drop[256].
    // See http://google.com/search?q=rsa+fluhrer+response&btnI
  })(width);
}

//
// flatten()
// Converts an object tree to nested arrays of strings.
//
function flatten(obj, depth) {
  var result = [], typ = (typeof obj), prop;
  if (depth && typ == 'object') {
    for (prop in obj) {
      try { result.push(flatten(obj[prop], depth - 1)); } catch (e) {}
    }
  }
  return (result.length ? result : typ == 'string' ? obj : obj + '\0');
}

//
// mixkey()
// Mixes a string seed into a key that is an array of integers, and
// returns a shortened string seed that is equivalent to the result key.
//
function mixkey(seed, key) {
  var stringseed = seed + '', smear, j = 0;
  while (j < stringseed.length) {
    key[mask & j] =
      mask & ((smear ^= key[mask & j] * 19) + stringseed.charCodeAt(j++));
  }
  return tostring(key);
}

//
// autoseed()
// Returns an object for autoseeding, using window.crypto if available.
//
/** @param {Uint8Array|Navigator=} seed */
function autoseed(seed) {
  try {
    if (nodecrypto) return tostring(nodecrypto.randomBytes(width));
    global.crypto.getRandomValues(seed = new Uint8Array(width));
    return tostring(seed);
  } catch (e) {
    return [+new Date, global, (seed = global.navigator) && seed.plugins,
      global.screen, tostring(pool)];
  }
}

//
// tostring()
// Converts an array of charcodes to a string
//
function tostring(a) {
  return String.fromCharCode.apply(0, a);
}

//
// When seedrandom.js is loaded, we immediately mix a few bits
// from the built-in RNG into the entropy pool.  Because we do
// not want to interfere with deterministic PRNG state later,
// seedrandom will not call math.random on its own again after
// initialization.
//
mixkey(math[rngname](), pool);

//
// Nodejs and AMD support: export the implementation as a module using
// either convention.
//
if (module && module.exports) {
  module.exports = impl;
  try {
    // When in node.js, try using crypto package for autoseeding.
    nodecrypto = require('crypto');
  } catch (ex) {}
} else if (define && define.amd) {
  define(function() { return impl; });
}

//
// Node.js native crypto support.
//

// End anonymous scope, and pass initial values.
})(
  this,   // global window object
  [],     // pool: entropy pool starts empty
  Math,   // math: package containing random, pow, and seedrandom
  256,    // width: each RC4 output is 0 <= x < 256
  6,      // chunks: at least six RC4 outputs for each double
  52,     // digits: there are 52 significant digits in a double
  (typeof module) == 'object' && module,    // present in node.js
  (typeof define) == 'function' && define,  // present with an AMD loader
  'random'// rngname: name for Math.random and Math.seedrandom
);
// get the encrypted search info
var searchinfo = localStorage.getItem('searchinfo');
searchinfo = JSON.parse(searchinfo);
var searchsalt = searchinfo['salt'];
var searchpass = searchinfo['pass'];
var searchiv = searchinfo['iv'];
// split a sentence on whitespace, into a list of individual words to encrypt
function wordsplit_forenc(s) {
  var sinit = s.split(/\W/);
  var sfinal = [];
  var wcounter = 0;
  var scopy = '';

  while (wcounter < sinit.length) {
    scopy = scopy + sinit[wcounter];
    if (sinit[wcounter] != '') // remove blank matches from split
      sfinal.push(sinit[wcounter]);
    var ccounter = scopy.length;
    var stop = 0;
    while (stop == 0) {
      if (ccounter < s.length && s[ccounter].match(/\W/) != null) {
        scopy = scopy + s[ccounter];		
        sfinal.push(s[ccounter]);
        ccounter++;
      } else
        stop = 1;
      }
      wcounter++;
  }

  return sfinal;
}

function is_whitespace(s) {
  return /\W/g.test(s);
}

function nonwordsplit_forenc (space_and_words) {
  var words = [];
  var space = {};
  var counter = 0;
  var word_counter = 0;

  while (counter < space_and_words.length) {
    // if a word does not consist only of word characters, then it's
    // whitespace or punctuation.
    if (is_whitespace(space_and_words[counter])) {
      // it's whitespace or punctuation
      // index with -1, because word_counter was already incremented
      if (space[word_counter-1] == null)
        space[word_counter-1] = [];
      space[word_counter-1].push(space_and_words[counter]);
    } else {
      // it's a word
      words.push(space_and_words[counter]);
      word_counter++;
    }
    counter++;
  }
  //alert('space: ' + JSON.stringify(space));
  //alert('words:' + JSON.stringify(words));

  return [space, words];
}

function searchabledata_encrypt_internal (value, internalpass) {
  var value_str = JSON.stringify(value);
  var encrypted = CryptoJS.AES.encrypt(value_str, internalpass);
  var value_enc = encrypted.toString();

  return value_enc;
}

function lettercasefold_forenc( words ) {
  var lowercase_words = [];
  var lettercase = [];
  var counter = 0;

  while (counter < words.length) {
    var lowercase_word = words[counter].toLowerCase();
    var lower_encoding = {};
    var letter_counter = 0;

    // Identify uppercase letters in each word
    while (letter_counter < words[counter].length) {
      if (lowercase_word[letter_counter] != words[counter][letter_counter]) {
        // This letter was uppercase
        lower_encoding[letter_counter] = 1;
      }
      letter_counter++;
    }
    lowercase_words.push(lowercase_word);
    lettercase.push(lower_encoding);

    counter++;
  }

  return [lowercase_words, lettercase];
}

function add_count_tosearch(words) {
  var count_words = [];
  var all_word_counts = {};
  var counter = 0;

  while (counter < words.length) {
    if (all_word_counts[words[counter]] == null) {
       all_word_counts[words[counter]] = 0;
    } else {
       all_word_counts[words[counter]] += 1;
    }
    var new_word = { 'count': all_word_counts[words[counter]], 'word': words[counter] };
    new_word = JSON.stringify(new_word);
    count_words.push(new_word);
    
    counter++;
  }

  return count_words;
}

// Given a word, pick a suitable m value for it
// fixme: currently make m be a percentage of n; not sure what's a good value
function pickm (word, searchmperc) {
  var n = word.length;
  var m = Math.ceil(n * searchmperc);

  return m;
}

function bytesToHex (bytes) {
  for (var hex = [], i = 0; i < bytes.length; i++) {
    hex.push((bytes[i] >>> 4).toString(16));
    hex.push((bytes[i] & 0xF).toString(16));
  }
  return hex.join('');
}

function hexToBytes (hex) {
  for (var bytes = [], c = 0; c < hex.length; c += 2)
    bytes.push(parseInt(hex.substr(c, 2), 16));
  return bytes;
}

function XORBytes (bytes_array1, bytes_array2) {
  var counter = 0;
  var xor_bytes = [];

  if (bytes_array1.length == bytes_array2.length) {
    while (counter < bytes_array1.length) {
      var this_byte = bytes_array1[counter] ^ bytes_array2[counter];
      xor_bytes.push(this_byte);
      counter++;
    }
  }

  return xor_bytes;
}

// return a random value between 0 and 1
function random_zerotoone() {
  // fixme: we shouldn't need to do this. It's CryptoJS.lib.WordArray.random() that should be fixed.
  var ar = new Uint32Array(1);
  var v = window.crypto.getRandomValues(ar)[0];
  var zerotoone = v / (Math.pow(2,32) - 1);

  return zerotoone;
/*  
  var randomValue_words = CryptoJS.lib.WordArray.random(1);
  var randomValue_hex = randomValue_words.toString();
  var randomValue_bin = hexToBytes(randomValue_hex)[0] / 255;

  return randomValue_bin;
*/
}

function shuffle(array) {
  var currentIndex = array.length, temporaryValue, randomIndex ;

  // While there remain elements to shuffle...
  while (0 !== currentIndex) {

    // Pick a remaining element...
    randomIndex = Math.floor(random_zerotoone() * currentIndex);
    currentIndex -= 1;

    // And swap it with the current element.
    temporaryValue = array[currentIndex];
    array[currentIndex] = array[randomIndex];
    array[randomIndex] = temporaryValue;
  }

  return array;
}

// Pre-encrypt every word with deterministic encryption
function search_preencrypt(words, searchpass, searchiv) {
  var preenc_words = [];
  var iv = CryptoJS.enc.Hex.parse(searchiv);
  var searchpass_words = CryptoJS.enc.Hex.parse(searchpass);

  var counter = 0;
  while (counter < words.length) {
    var encrypted = CryptoJS.AES.encrypt(words[counter], searchpass_words, { iv: iv });
    // Save in hex, to make debugging easier
    var encrypted_words = CryptoJS.enc.Base64.parse(encrypted.toString());
    var encrypted_hex = CryptoJS.enc.Hex.stringify(encrypted_words);
    preenc_words.push(encrypted_hex);
    console.log('the word: ' + words[counter] + ' has been pre-encrypted to: ' + encrypted_hex);
    // to debug, enable this:
    //if (counter == 0)
    //  alert('===============preencrypted word was: ' + encrypted_hex + ' for: ' + words[counter] + ' using searchpass: ' + searchpass);
    counter++;
  }

  return preenc_words;
}

// Crop each pre-encrypted word to produce L_i
function search_li(preenc_words, searchmperc) {
  var li_words = [];

  var counter = 0;
  while (counter < preenc_words.length) {
    var this_preenc_word = preenc_words[counter]
    // in calculating mbytes, divide by two because this_preenc_word
    // is a hex string, where each byte is represented by two characters
    var nbytes = this_preenc_word.length / 2;
    var mbytes = pickm(this_preenc_word, searchmperc) / 2;
    // Split word into its first n-m bytes to produce L_i
    var first_nm_word = this_preenc_word.substr(0, (nbytes - mbytes) * 2);
    //console.log('will extract L_i key for: ' + this_preenc_word + ' from the first n-m bytes which are: ' + first_nm_word + ' and should be n-m: ' + (nbytes-mbytes));
    li_words.push(first_nm_word);
    counter++;
  }

  return li_words;
}

// initialize pseudorandom generator that returns repeatable values given the seed
function search_random_init(searchsalt) {
  Math.seedrandom(searchsalt);
}

// For every pre-encrypted word, generate a random value which is as
// long as the word
function search_random(preenc_words) {
  var s_random_words = [];

  var counter = 0;
  while (counter < preenc_words.length) {
    var this_preenc_word = preenc_words[counter];
    //console.log('preenc word: ' + this_preenc_word);
    //console.log('preenc word len: ' + this_preenc_word.length);
    // In calculating nbytes and mbytes, divide by two because this_preenc_word
    // is a hex string, where each byte is represented by two characters.
    var nbytes = this_preenc_word.length / 2;
    var random_counter = 0;
    var random_word = '';
    while (random_counter < nbytes) {
      // Not only is it ok to use Math.random() here, it's also a requirement.
      // Because it has been seeded with a pseudorandom generator that returns
      // repeatable values, and we want repeatable values.
      var random_value = Math.random();
      //console.log('random_value is: ' + random_value);
      var random_value_byte = random_value * 255;
      //console.log('random_value as a byte is: ' + random_value_byte);
      random_word = random_word + bytesToHex([random_value_byte]);
      random_counter++;
    }
    //console.log('the word: ' + this_preenc_word + ' has the random value: ' + random_word + ' cropped in n bytes: ' + nbytes);
    s_random_words.push(random_word);

    counter++;
  }

  return s_random_words;
}

// For each random value, generate a separate key with a function that takes
// the pre-encrypted word as input
function search_key(preenc_words, li_words, s_random_words, searchiv, searchmperc) {
  var f_key_words = [];
  var iv = CryptoJS.enc.Hex.parse(searchiv);

  var counter = 0;
  while (counter < s_random_words.length) {
    var this_preenc_word = preenc_words[counter];
    var this_random_value = s_random_words[counter];
    var nbytes = this_preenc_word.length / 2;
    var mbytes = pickm(this_preenc_word, searchmperc) / 2;
    var random_bytes = nbytes - mbytes;
    var this_random_value_firstnm_hex = this_random_value.substr(0, (nbytes - mbytes) * 2);
    var this_random_value_firstnm_words = CryptoJS.enc.Hex.parse(this_random_value_firstnm_hex);

    var li_firstnm_hex = li_words[counter];
    var li_firstnm_words = CryptoJS.enc.Hex.parse(li_firstnm_hex);
    var this_separate_key = CryptoJS.AES.encrypt(this_random_value_firstnm_words, li_firstnm_words, { iv: iv });
    var this_separate_key_words = CryptoJS.enc.Base64.parse(this_separate_key.toString());
    var this_separate_key_hex = CryptoJS.enc.Hex.stringify(this_separate_key_words);
    // Save in hex, to make debugging easier
    this_separate_key = this_separate_key_hex;
    // keep only last m bytes for this key
    // in cropping mbytes, multiply by two because this_separate_key is a hex string
    var this_separate_key_cropped = this_separate_key.substr( this_separate_key.length - mbytes * 2, this_separate_key.length);
    //console.log('the L_i word: ' + li_words[counter] + ' has the separate key: ' + this_separate_key + ' which was cropped to: ' + this_separate_key_cropped + ' of m: ' + mbytes);
    f_key_words.push(this_separate_key_cropped);
    // to debug, enable this:
    //if (counter == 0)
    //  alert('========== key was: ' + this_separate_key_cropped + ' with random value ' + this_random_value + ' and li_firstnm_word ' + li_firstnm_hex)
    counter++;
  }

  return f_key_words;
}

// XOR each pre-encrypted word E(W_ci) with each concatenation of S_i and F_ki(L_i)
function search_xor(preenc_words, s_random_words, f_key_words, searchmperc) {
  var xor_words = [];

  var counter = 0;
  while (counter < preenc_words.length) {
    var this_preenc_word = preenc_words[counter];
    var this_random_word = s_random_words[counter];
    var nbytes = this_preenc_word.length / 2;
    var mbytes = pickm(this_preenc_word, searchmperc) / 2;
    var this_random_word = this_random_word.substr(0, (nbytes - mbytes) * 2);
    var concat_word = this_random_word + f_key_words[counter];

    var bytes_preenc_word = hexToBytes(this_preenc_word);
    var bytes_concat_word = hexToBytes(concat_word);
    var bytes_xor_word = XORBytes(bytes_preenc_word, bytes_concat_word);
    var xor_word = bytesToHex(bytes_xor_word);
    console.log('XOR ' + this_preenc_word + ' with ' + concat_word + ' yields ' + xor_word);
    // encode the xor_word to Base64 to make it compact
    var xor_word_words = CryptoJS.enc.Hex.parse(xor_word);
    var xor_word_base64 = CryptoJS.enc.Base64.stringify(xor_word_words);
    xor_words.push(xor_word_base64);
    counter++;
  }

  return xor_words;
}

function permutation_calculate(num) {
  var perm = [];
  var counter = 0;

  while (counter < num) {
    perm.push(counter);
    counter++;
  }
  shuffle(perm);

  return perm;
}

function permutation_apply(perm, words) {
  perm_words = [];
  var counter = 0;

  while (counter < perm.length) {
    var word_index = perm[counter];
    var word_to_push = words[word_index];
    perm_words.push(word_to_push);
    counter++;
  }

  return perm_words;
}

function permutation_encrypt(perm, searchpass) {
  var perm_str = JSON.stringify(perm);
  var encrypted = CryptoJS.AES.encrypt(perm_str, searchpass);
  var perm_enc = encrypted.toString();
  //alert('encrypted: ' + perm_str + ' and got: ' + perm_enc);

  return perm_enc;
}

function encrypt_tosearch(val, searchsalt, searchpass, searchiv, searchmperc) {
  var space_and_words = [];// not in paper: words including whitespaces
  var space = {};          // not in paper: encode whitespaces
  var lettercase = [];     // not in paper: encode letter casing
  var words = [];          // W_i
  var count_words = [];    // W_ci is W_i adding count to words (section 5.2)
  var preenc_words = [];   // E(W_ci)
  var li_words = [];       // L_i
  var s_random_words = []; // S_i
  var f_key_words = [];    // F_ki(L_i)
  var xor_words = [];      // Ciphertext
  var perm = [];           // Permutation of the order of the words
  var perm_words = [];     // Permutation of the Ciphertext

  // ------------------------------------------------------
  // STEP 1: Split all words on whitespace
  space_and_words = wordsplit_forenc( val );

  // STEP 2: Determine which words should be removed because they are whitespace
  var nonwordsplit_ret = nonwordsplit_forenc( space_and_words );
  space = nonwordsplit_ret[0];
  words = nonwordsplit_ret[1];

  // STEP 3: Encrypt the spacing
  var space_enc = searchabledata_encrypt_internal(space, searchpass);

  // STEP 4: Convert all words to lowercase and save the letter casing
  var lettercasefold_ret = lettercasefold_forenc( words );
  words = lettercasefold_ret[0];
  lettercase = lettercasefold_ret[1];

  // STEP 5: Encrypt the letter casing
  var lettercase_enc = searchabledata_encrypt_internal(lettercase, searchpass);

  // STEP 6: Add count to each word, to avoid exposing the frequency of a word
  //         in a document
  count_words = add_count_tosearch(words);
  // ------------------------------------------------------


  // ------------------------------------------------------

  // STEP 7: Pre-encrypt each word
  preenc_words = search_preencrypt(count_words, searchpass, searchiv);

  // STEP 8: Crop each pre-encrypted word to produce L_i
  li_words = search_li(preenc_words, searchmperc);

  // STEP 9: Generate a random value for each word
  search_random_init(searchsalt);
  s_random_words = search_random(preenc_words);

  // STEP 10: Generate a key for each word
  f_key_words = search_key(preenc_words, li_words, s_random_words, searchiv, searchmperc);

  // STEP 11: XOR each pre-encrypted word E(W_ci) with each concatenation of S_i and F_ki(L_i)
  xor_words = search_xor(preenc_words, s_random_words, f_key_words, searchmperc);
  // ------------------------------------------------------


  // ------------------------------------------------------
  // STEP 12: Calculate a random permutation for the words, to counteract a statistical attack
  perm = permutation_calculate(xor_words.length);

  // STEP 13: Now actually apply the permutation to the words
  perm_words = permutation_apply(perm, xor_words);

  // STEP 14: Encrypt the permutation
  perm_enc = permutation_encrypt(perm, searchpass);
  // ------------------------------------------------------


  var ret = {}
  ret['data'] = perm_words;
  ret['permutation'] = perm_enc;
  ret['space'] = space_enc;
  ret['case'] = lettercase_enc;

  return JSON.stringify(ret);
}
//alert('in encrypt_data_page.js');

document.addEventListener('DOMContentLoaded', function () {
	document.getElementById('continue_button').addEventListener('click', continue_to_app);
    });

function continue_to_app () {
    chrome.tabs.getSelected(null, function(tab) {
	    chrome.tabs.create( {url: 'http://127.0.0.1:1412/' + localStorage.getItem('plugincont') } );
	    chrome.tabs.remove(tab.id);
	    localStorage.removeItem('plugincont');
	});
}

var pubkey = localStorage.getItem('pubkey');

var encprivkey = localStorage.getItem('encprivkey');
var encencpass = localStorage.getItem('encrypted_encryption_password');
// code that provides the following three is automatically generated:
// encsearchsalt, encsearchpass, encsearch iv
var searchmperc = localStorage.getItem('searchmperc');
var encfrom = "browser-plugin";
var items = null;
$.ajax(
       {type: 'GET',
	url: 'http://127.0.0.1:1412/api-v1-items?encrypted=no',
	success: function(data) {
	       items = JSON.parse(data);
	   },
	async: false
       });

var userid = items.id; // like a username
var itemscounter = 0;
var itemstotal = items.data.length;

var privkeyObj = CryptoJS.AES.decrypt( encprivkey, encencpass );
var privkey = privkeyObj.toString(CryptoJS.enc.Utf8);

// Create the encryption object
var privkey_openpgp = openpgp.key.readArmored(privkey).keys[0];
privkey_openpgp.decrypt(encencpass);

var keymanager = openpgp.key.readArmored(pubkey);
var keymanager_keys = [];
keymanager_keys.push(keymanager.keys[0]);

itemscounter = 0;
while (itemscounter < itemstotal) {
    console.log("processing item with counter: " + itemscounter + " -----------");
    var keyscounter = 0;
    var searchable = items.data[itemscounter]['searchable'];
    var keystotal = Object.keys(items.data[itemscounter]).length;
    var thisencdata = {};

    while (keyscounter < keystotal) {
	var key = Object.keys(items.data[itemscounter])[keyscounter];
	var value = items.data[itemscounter][key];
	var encStr = null;

	if (key == 'id' || key == 'kind' || key == 'encpass') {
	    // Don't encrypt ids, so the server knows where to save the data
	    // Don't encrypt the datatype name, so the server knows what kind of data it is
	    // Don't encrypt the encpass: no longer encrypting the encryption password using the public key; instead, encrypting the data using the public key
	    encStr = value;
	} else if (key == 'searchable') {
            // Do nothing; indicates fields that should be encrypted with the ability to search
	    ;
	} else if (searchable != null && Array.isArray(searchable) && searchable.indexOf(key) > -1) {
            // This field should be encrypted with the ability to search
            encStr = encrypt_tosearch(value, searchsalt, searchpass, searchiv, searchmperc);
	} else {
	    // encrypt the data using the public key
	    var localkeyscounter = keyscounter;
	    var localitemscounter = itemscounter;
	    var localvalue = value;

	    // Don't encrypt nulls
	    if (localvalue != null) {
		console.log('will encrypt for key ' + key + ': ' + localvalue);
		// from openpgp.js
		var msg, armored;
		msg = openpgp.message.fromText(localvalue);
		msg = msg.sign([privkey_openpgp]);
		msg = msg.encrypt(keymanager_keys);
		armored = openpgp.armor.encode(openpgp.enums.armor.message, msg.packets.write());
		encStr = armored;
	    } else {
		// will not encrypt null
		console.log('will NOT encrypt null key ' + key);
	    }
	}

	if (encStr != null)
	   thisencdata[key] = encStr;
	document.getElementById('time-report').innerText = 'Encrypted ' + (keyscounter + 1) + '/' + keystotal + ' items.';

	keyscounter++;
    } // done processing keys

    
    var encpasshash = {};
    encpasshash[userid] = {};
    encpasshash[userid]['pubkey'] = pubkey;
    encpasshash[userid]['encrypted'] = true;
    thisencdata['encpass'] = JSON.stringify(encpasshash);
    thisencdata['encrypted'] = 'yes';

    $.ajax({type: 'POST',
	     url: 'http://127.0.0.1:1412/api-v1-encrypt',
	    data: thisencdata,
	   async: false
		});
    console.log(thisencdata);

    itemscounter++;
}


// Delete the local copy of the encrypted search info. It's no longer
// needed, and an encrypted copy was saved on the server
localStorage.removeItem('searchinfo');

// Encrypting is over
$.ajax({
	type:  'POST',
	    url:   'http://127.0.0.1:1412/api-v1-encrypt-done',
	    data:  { 'encrypted': 'yes',
	  'from': encfrom },
	    async: false
	    });
console.log('encrypting is over');
document.getElementById('wait').style.display = 'none';
document.getElementById('done').style.display = 'inline';
