(defop example-contact-add req
  (page (backlink "/examples") site-name* nil
	(encverif (get-user req))
	(pr "This example shows how to add contacts. It shows how you can verify a contact is who you think it is, before sending the contact an encrypted message.")
	(br2)
	(pr "Contact names are encrypted in a way that allows searching for them. This is useful when using autocomplete to find the contact you want to send a message to.")
	(br2)
	(pr "Although a username or email address stay unencrypted, the name is encrypted in a way that can be searched. This obviously offers no protection if the username or email address contain a person's name. But that's not a limitation of " site-name* ". It's a limitation in the design of the email protocol, because email addresses are always disclosed to mail servers. One could build a better protocol that separates address from identity, and does not reveal to the server people's names.")
	(br2)
	(pr "A contact must have a public key. The public key can be manually entered by the user, or it can be sent by the server: a server can use a public key directory or manage public keys itself.")
	(br2)
	(pr "It's up to the user to verify the fingerprint of the contact's public key: that the key belongs to the person the user thinks it does. To some degree the plugin helps. When a contact is created, the plugin displays the fingerprint and asks the user to confirm it.")
	(br2)
	(pr "After verification, a mapping between the key id of the public key and the contact id (like a username) are encrypted and saved in the browser plugin.")
	(br2)
	(pr "When encrypting messages, the browser plugin verifies the recipient's public key hasn't changed to protect from man-in-the-middle attacks. Although the recipient's public key is retrieved from the server, the plugin verifies the key id of the key matches the key id that was saved in the plugin.")
	(br2)
	;; I think this was removed
	;; (pr "There is an additional check: that the contact id hasn't changed. This detects if the contact switched to a different user account or uses more than one account with the same public key.")
	;; (br2)
	(tag (ul)
	     (tag (li)
		  (w/link (example-contact-add-core req "username")
			  (pr "Enter a username, server fetches the public key."))
		  (br2))
	     (tag (li)
		  (w/link (example-contact-add-core req "keyserver")
			  (pr "Enter an email, server fetches the key from a PGP keyserver."))
		  (br2))
	     (tag (li)
		  (w/link (example-contact-add-core req "pubkey")
			  (pr "Enter the public key."))
		  (br2)))
	(br)
	(click-source "joan-example-contact-add.arc")
	(br 8)
	(hline)
	))

(= enccontact* "gandalf")
(defop api-v1-pubkey req
  (withs (id (arg req "id") pubkey nil)
	 (if (is id enccontact*)
	     (= pubkey (fread "joan-example-contact-add-somefakepubkey")))
	 (pr "{ \"id\": \"" id "\""
	     ", \"pubkey\": " (if pubkey (tojson pubkey) "null")
	     "}")))

(def gpg-fetchkey (email)
  (withs (url (str "https://pgp.mit.edu/pks/lookup?search=" (urlencode email) "&op=get"))
	 (tostring (system (str "curl -s '" url "' | sed -e '1,/<pre>/d'  | sed -e '/pre>/,$d'")))))

(defop api-v1-pgp-pubkey req
  (withs (email (arg req "email"))
	 (if email
	     (pr (gpg-fetchkey email)))))

(def example-contact-add-core (req kind)
  (page (backlink "/examples") site-name* nil
	;; define what the browser extension should do
	(rex (list ;; you still need the encrypt command to process a
		   ;; contact, even if you won't encrypt anything
		   (list 'encrypt
			 (obj pubuserid (get-user req)
			      formbuttons (list (obj form "ENTER_TEXT"
						     btn 'example-text
						     encpassid "encpass"
						     contact-list (list (obj action 'add
									     id (if (is kind "keyserver")
										    "email"
										    "username")
									     fn "get-pubkey"
									     ;; NOTE: the plugin automatically adds a cantact's savepubkeyto field to the enc-itemlist
									     savepubkeyto "pubkey"
									     saveistextarea t))
						     ))
			      ;; Don't manually list pubkey here. It is encrypted automatically.
			      enc-itemlist (list (obj id "name"
						      tosearch t
						      tosearchout "nameout"))
			      ))
		   )
	     ;; execute code written on the server
	     (freadrep "joan-example-contact-add-sent.arc"
		       (list (list "@KIND@" kind))))
	(if (is kind "username")
	    (pr "Enter the username '" enccontact* "', press 'submit', and the browser plugin will fetch the user's public key from the server and ask you to verify it.")
	    (is kind "keyserver")
	    (pr "Enter an email address, press 'submit', and the browser plugin will fetch the user's public key from a PGP keyserver, through the server, and ask you to verify it.")
	    (is kind "pubkey")
	    (pr "Enter a username and a public key, press 'submit', and the browser plugin will use the public key you entered and ask you to verify it."))
	(br2)
	(pr "Click ")
	(tag (a href (str "/joan-example-contact-add-sent.arc"))
	     (pr "here"))
	(pr " to see the code written on the server that is executed by the browser extension.")
	(br 3)
	(sform "ENTER_TEXT"
	       (fn (_)
		   (withs (p (prof (get-user _))
			   encverifiedkeyids (arg _ "encverifiedkeyids")
			   encpass (arg _ "encpass")
			   username (arg _ "username")
			   pubkey (arg _ "pubkey")
			   nameout (arg _ "nameout")
			   numwords 0)
			  (= nameout (fromjson nameout))
			  (fwrite "/tmp/nameout" (tostring (pr nameout)))
			  (= p!encverifiedkeyids encverifiedkeyids)
			  (save-prof p)
			  (if (is kind "keyserver")
			      (= username (arg _ "email")))
			  ;; save the contact
			  (add-contact username p!user pubkey
				       (nameout "data") (nameout "permutation") (nameout "space") (nameout "case"))
			  ;; the server is expected to keep track of
			  ;; the maximum words in a contact name, to tell
			  ;; the client to supply at least as many
			  ;; keys (one per word in the name)
			  (= numwords (len (nameout "data")))
			  (if (or (no p!searchmaxnames) (> numwords p!searchmaxnames))
			      (do (= p!searchmaxnames numwords)
				  (save-prof p)))
			  (example-contacts-show p!user)))
	       ;; each data item has an encryption password
	       (encverif (get-user req))
	       (encsearchinfo (get-user req))
	       (gentag input type "hidden" id "encpass" name "encpass" value "")
	       (gentag input type "hidden" id "nameout" name "nameout" value "")
	       (tag (table)
		    (if (is kind "keyserver")
			(tr (tag (td) (pr "Email: "))
			    (tag (td) (gentag input type 'text id "email" name "email" style "width:225px;" value "laura.poitras@theintercept.com" autofocus "")))
			(tr (tag (td) (pr "Username: "))
			    (tag (td) (gentag input type 'text id "username" name "username" style "width:165px;" value enccontact* autofocus ""))))
		    (if (is kind "pubkey")
			(tr
			 (tag (td style "vertical-align:top;")
			      (pr "Public key: "))
			 (tag (td)
			      (tag (textarea id "pubkey" name "pubkey" rows 8 cols 42)
				   (pr "")))))
		    (tr
		     (tag (td)
			  (pr "Name: "))
		     (tag (td)
			  (gentag input type 'text id "name" name "name" style "width:165px;"
				  value (if (is kind "keyserver")
					    "Laura Poitras"
					    "")
				  )))
		    )
	       ;; Create a pubkey field regardless, because that's
	       ;; where savepubkeyto will save the public key after
	       ;; encrypting it
	       (if (or (is kind "username")
		       (is kind "keyserver"))
		   (gentag input type 'hidden id "pubkey" name "pubkey"))
	       ;; the plugin enforces signing a pubkey that will be
	       ;; encrypted for a contact even if signing isn't
	       ;; enabled by the user
	       ;;
	       ;; BUT, whenever anything else is encrypted, we get a
	       ;; signing warning; so force signing to disable it
	       (showsign nil t)
	       
	       (br 3)
	       (tag (a id "example-text")
		    (pr "submit")))
	(br 8)
	(hline)))
